//
// Text Data Structure
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 10.28.98 $
//


#include "st.h"
#include "dyna.h"

//#define DEBUG_DYNA

void Text::validate(void) { // can be optimized passing a map_tree_iterator
  //
  // validate the tree
  //
 map_tree
   temp_tree (tree);
 int
   i = 0;
  tree.erase(tree.begin(), tree.end());
  for (map_tree_iterator mt = temp_tree.begin(); mt != temp_tree.end(); ++mt) {
    tree[i + mt->second.len - 1] = node_type(i, mt->second.len, mt->second.begin, mt->second.end);
    i += mt->second.len;
  }
}


void Text::defragment(void) {
  //
  // defragment the tree: merge all chunks into a single vector
  //
  if (tree.size() > 1) {
    text_vector
      temp_text(text);
    text.erase(text.begin(), text.end());
    for (map_tree_iterator mt = tree.begin(); mt != tree.end(); ++mt)
      copy(mt->second.begin, mt->second.end, back_inserter(text));
    tree.erase(tree.begin(), tree.end());
    tree[text.size() - 1] = node_type(0, text.size(), text.begin(), text.end());
  }
}

  
void Text::erase (int i, int j) {
  //
  // erase all the symbols [i,j) from the data structure
  //
  if (j <= i)
    return;

  map_tree_iterator
    vi = tree.lower_bound(i),
    vj = tree.lower_bound(j);

  if (vi == tree.end())
    return;
  else if (vj == tree.end()) {
    text_vector_iterator
      ib = vi->second.begin;
    int
      lbi = vi->second.lb,
      new_isize = i - lbi;

    tree.erase(vi, tree.end());

    if (new_isize > 0)
      tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);

    } else {
      text_vector_iterator
	ib = vi->second.begin,
	je = vj->second.end;
      
      int
	lbi = vi->second.lb,
	lbj = vj->second.lb,
	new_isize = i - lbi,
	new_jsize = vj->second.len - (j - lbj);
      
      tree.erase(vi, ++vj);
      
      if ((new_jsize > 0) && (new_isize > 0)) {
	tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);
	tree[lbi + new_jsize + new_isize - 1] = node_type(lbi + new_isize, new_jsize, je - new_jsize, je);
      } else {
	if (new_jsize > 0)
	  tree[lbj + new_jsize - 1] = node_type(lbj, new_jsize, je - new_jsize, je);
	if (new_isize > 0)
	  tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);
      }
    }
  validate();
}



void Text::substitute (int i, int j,  text_vector& pattern) {
  //
  // substitute all the symbols in [i,j) with the pattern
  //
  if (j <= i)
    return;

# ifdef DEBUG_DYNA
  cout << " --> substitute " << i << " " << j << " with " << pattern << endl; 
# endif

  map_tree_iterator
    vi = tree.lower_bound(i),
    vj = tree.lower_bound(j);
  int
    pat_size = pattern.size();

  //cout << pat_size  << endl;
  //cout << i << " " << j << " " << j - i << endl << endl;

  assert (pat_size <= j - i);

  if (vi == tree.end())
    return;
  else if (vj == tree.end()) {
    text_vector_iterator
      ib = vi->second.begin;
    int
      lbi = vi->second.lb,
      new_isize = i - lbi + pat_size;

#   ifdef DEBUG_DYNA
    cout << " to the end, keys ki=" << vi->first << endl; 
    cout << " lbi=" << lbi << endl;
    cout << " newi=" << new_isize << endl;
#   endif

    tree.erase(vi, tree.end());
    copy(pattern.begin(), pattern.end(), ib + (i - lbi));
    
    if (new_isize > 0)
      tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);
    
  } else {
    text_vector_iterator
      ib = vi->second.begin,
      je = vj->second.end;
    
    int
      lbi = vi->second.lb,
      lbj = vj->second.lb,
      new_isize = i - lbi + pat_size,
      new_jsize = vj->second.len - (j - lbj);
    
#   ifdef DEBUG_DYNA
    cout << " keys ki=" << vi->first << " kj=" << vj->first ; 
    cout << " lbi=" << lbi << " lbj=" << lbj;
    cout << " newi=" << new_isize << " newj=" << new_jsize << endl;
#   endif
    
    tree.erase(vi, ++vj);
    copy(pattern.begin(), pattern.end(), ib + (i - lbi));
    
    if ((new_jsize > 0) && (new_isize > 0)) {
#     ifdef DEBUG_DYNA
      cout << " newkeys ki=" << (lbi + new_isize - 1) << " kj=" << (lbj + new_jsize + new_isize - 1) << endl; 
#     endif
      tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);
      tree[lbi + new_jsize + new_isize - 1] = node_type(lbi + new_isize, new_jsize, je - new_jsize, je);
    } else {
      if (new_jsize > 0) {
#       ifdef DEBUG_DYNA
	cout << " newkey kj=" << (lbj + new_jsize - 1) << endl; 
#       endif
	tree[lbj + new_jsize - 1] = node_type(lbj, new_jsize, je - new_jsize, je);
      }
      if (new_isize > 0) {
#       ifdef DEBUG_DYNA
	cout << " newkey ki=" << (lbi + new_isize - 1) << endl; 
#       endif
	tree[lbi + new_isize - 1] = node_type(lbi, new_isize, ib, ib + new_isize);
      }
    }
  }
  validate();
}



vector<int> Text::erase (text_vector& pattern) {
  //
  // search for all the occurrences of the pattern
  // if more than MIN_OCC, erase them from the data structure
  // returns a vector with the positions
  //
  vector<int>
    positions;
  tv_iterator
    tv = tree.begin(),
    tv_end = tree.end();
  int
    pos,
    del = 0;

  while ((tv = search(tv, tv_end, pattern.begin(), pattern.end())) != tree.end()) {
    positions.push_back(tv.get_pos());
    tv = tv + pattern.size();
  }

  if (positions.size() < MIN_OCC)
    return positions;

  for (vector<int>::iterator i = positions.begin(); i != positions.end(); ++i) {
    pos = *i - del;
    erase(pos, pos + pattern.size());
    del += pattern.size();
  }

  if ((int)(tree.size() * tree.size() * tree.size()) > size())
    defragment();

  return positions;
}



vector<int> Text::substitute (text_vector& pattern, text_vector& new_pattern) {
  //
  // search for all the occurrences of the pattern
  // if more than MIN_OCC, substitute them in the data structure with
  // the new pattern
  //
  vector<int>
    positions;
  tv_iterator
    tv = tree.begin(),
    tv_end = tree.end();
  int
    pos,
    del = 0;

  while ((tv = search(tv, tv_end, pattern.begin(), pattern.end())) != tree.end()) {
    positions.push_back(tv.get_pos());
    tv = tv + pattern.size();
  }

  if (positions.size() < MIN_OCC)
    return positions;

  for (vector<int>::iterator i = positions.begin(); i != positions.end(); ++i) {
    pos = *i - del;
    substitute(pos, pos + pattern.size(), new_pattern);
#   ifdef DEBUG_DYNA
    cout << "Text Fragments " << endl << *this << endl;
#   endif
    del += pattern.size() - new_pattern.size();
  }

  if ((int)(tree.size() * tree.size() * tree.size()) > size())
    defragment();

  return positions;
}

#ifdef OFF3
vector<int> Text::substitute2 (text_vector& pattern) {
  //
  // search for all the occurrences of the pattern
  // if more than MIN_OCC, substitute them (except one) in the data structure with
  // the new pointer
  //
  vector<int>
    positions;
  tv_iterator
    tv = tree.begin(),
    tv_end = tree.end();
  int
    pos,
    del = 0;

  while ((tv = search(tv, tv_end, pattern.begin(), pattern.end())) != tree.end()) {
    positions.push_back(tv.get_pos());
    tv = tv + pattern.size();
  }

  if (positions.size() < MIN_OCC)
    return positions;

  int
    middle = positions.size() / 2,
    anchor = positions[middle];

  for (vector<int>::iterator i = positions.begin(); i != positions.end(); ++i)
    if (*i != anchor) {
      text_vector
	ptr;
      ptr.push_back(symbol((unsigned char)(anchor % 256), POINTER));
      ptr.push_back(symbol((unsigned char)(anchor / 256), POINTER));
      ptr.push_back(symbol((unsigned char)((pattern.size()-MIN_LENGTH) % 256), POINTER));
      pos = *i - del;
      substitute(pos, pos + pattern.size(), ptr);
#     ifdef DEBUG_DYNA
      cout << "Text Fragments " << endl << *this << endl;
#     endif
      del += pattern.size() - ptr.size();
    }

  if ((int)(tree.size() * tree.size() * tree.size()) > size())
    defragment();

  return positions;
}
#endif
