//
// Text Data Structure
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 10.28.98 $
//

#if !defined( _DYNA_H )
#define _DYNA_H

//#define DEBUG_DYNA

#include "st.h"
#include <map.h>
#include <vector.h>

class node_type {
public:
  int
    lb, // lower bound
    len; // invariant: lb + len == key - 1
  text_vector_iterator
    begin,
    end;
  // ----- Constructors
  node_type() :
    lb(0), len(0), begin(NULL), end(NULL) {}

  node_type(int l, int le, const text_vector_iterator b, const text_vector_iterator e) :
    lb(l), len(le), begin(b), end(e) {}
};

typedef map<int,node_type,less<int> > map_tree;
typedef map_tree::iterator map_tree_iterator;

class tv_iterator : public map_tree_iterator  {
protected:
  map_tree_iterator
    node;             // pointer to the node
  int
    pos;              // displacement
public:
  // ----- Constructor
  tv_iterator() {}
  tv_iterator(const map_tree_iterator& x) :  node(x), pos(0) {}

  // ----- Operators
  inline symbol operator*() {
    return *(node->second.begin + pos);
  }

  inline tv_iterator& operator++() {
    ++pos;
    if (pos >= node->second.len) {
      ++node;
      pos = 0;
    }
    return *this;
  }

  inline tv_iterator& operator--() {
    --pos;
    if (pos < 0) {
      --node;
      pos = node->second.len - 1;
    }
    return *this;
  }

  inline tv_iterator& operator+(unsigned int n) {
    pos += n;
    while (pos > node->second.len) { 
      pos -= node->second.len;
      ++node;
    }
    if (pos == node->second.len) {
      pos = 0;
      ++node;
    }
    return *this;
  }

  /* inline tv_iterator& operator-(int n) {
    pos -= n;
    while ((node != NULL)&&(pos < 0 ) {
      --node;
      pos += node->second.len;
    }
    return *this;
  } */

  inline bool operator==(const tv_iterator& x) {
    return (node == x.node) && (pos == x.pos);
  }

  inline bool operator!=(const tv_iterator& x) {
    return (node != x.node) || (pos != x.pos);
  }

  inline bool operator==(const map_tree_iterator& x) {
    return node == x;
  }

  inline bool operator!=(const map_tree_iterator& x) {
    return node != x;
  }

  friend ostream& operator<< (ostream& st, tv_iterator x) {
    st << "[key=" << (x.node)->first;
    st << ", lb=" << (x.node)->second.lb;
    st << ", len=" << (x.node)->second.len;
    st << ", pos=" << x.pos;
    st << ", ptr_to=" << (x.node)->second.begin + x.pos;
    st << "]" << endl;
    return st;
  }

  inline int get_pos() {
    return node->second.lb + pos;
  }
};


class Text {

private:
  map_tree
    tree;

public:
  text_vector
    text;
  // ----- Constructors
  Text() :
    tree() {}

# ifdef OFF3
  Text(char_vector& t) :
    tree(), text() {
    text.reserve(t.size());
    for (char_vector_iterator i = t.begin(); i != t.end(); ++i)
      text.push_back(symbol(*i));
    tree[text.size() - 1] = node_type(0, text.size(), text.begin(), text.end());
  }
# endif

  Text(text_vector& t) :
    tree(), text(t)
    {tree[t.size() - 1] = node_type(0, t.size(), t.begin(), t.end());}

  // ----- Outputs
  friend ostream& operator<< (ostream& st, Text l) {
    for (map_tree_iterator mt = l.tree.begin(); mt != l.tree.end(); ++mt) {
      st << "<";
      for (text_vector_iterator i = mt->second.begin; i != mt->second.end; ++i)
	st << *i;
      st << ">[" << mt->second.lb << ", k=" << mt->first << ", size=" <<mt->second.len << "] ";
    }
    return st;
  }

  // ----- Size
  int size(void) {
    return 1 + (--tree.end())->first;
  }

  // ----- Accessor
  symbol operator[](int i) {
    assert (tree.lower_bound(i) != tree.end());
    return (*((*tree.lower_bound(i)).second.begin + (i - (*tree.lower_bound(i)).second.lb)));
  }

  // ----- Accessor
  symbol at(int i) {
    //
    // retrieve the char at position i
    //
    map_tree_iterator
      vp = tree.lower_bound(i);
    if (vp == tree.end())
#ifdef OFF3
      return symbol(0,TRUE);
#else
      return (char)0;
#endif
    else
      return (*(vp->second.begin + (i - vp->second.lb)));
  }

  // ----- Other
  void validate(void);
  void defragment(void);
  void erase (int, int);
  void substitute (int, int, text_vector&);
  vector<int> erase (text_vector&);
  vector<int> substitute (text_vector&, text_vector&);
  vector<int> substitute2 (text_vector&);
  char_vector to_char_vector (void);
  char_vector to_char_vector_no_ptr (void);
};

#endif
