//
// Statistics Construction
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 10.28.98 $
//

#include <assert.h>
#include <fstream.h>
#include "st.h"
#include "dyna.h"
//#include "pm.h"
#include "zlibengn.h"
#include "arith.h"

#define AUGMENTED
// #define FINAL_ENCODING // recompute the final encoding at each substitution
// #define SAVEENCODING
// #define SAVECOMPRESSEDENCODING
// #define DEBUGFINALTREE
// #define PRINTTEXT
// #define PRINTTREE
// #define PRINTQUEUE
#define PRINTENCODER
// #define DEBUGSTAT
// #define DEBUG
// #define VERBOSE // print all the step performed
#define MAX_256 // does only 256 iterations

double
  hTEXT,
  hTXT,
  hOCC,
  hLEN,
  hPOS;
base2
  loga;

#ifdef OFF3
char_vector
  ttext,
# ifndef MAX_256
  ttext_h,
# endif
  ddict,
  sublen_vec,
  delta;
#else
char_vector
  ttext,
  sublen_vec,
  dict_vec,
  occurr_vec,
  abspol_vec,
  abspoh_vec,
  relpol_vec,
  relpoh_vec; 
#endif

extern void start_time2 (void);
extern void end_time2 (void);


void print_vec (char_vector& x) {
  //
  // print a vector<char> in numbers
  //
  for (char_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << " " << (unsigned short)(*i & 0xFF);
}



void print_text_vec (text_vector& x) {
  //
  // print a vector<char> in chars
  //
  for (text_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << *i;
}



void print_vec_char (char_vector& x) {
  //
  // print a vector<char> in chars
  //
  for (char_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << *i;
}



void save_encoding(string fn, char_vector& v1, char_vector& v2,
		   char_vector& v3, char_vector& v4, char_vector& v5,
		   char_vector& v6, char_vector& v7, char_vector& v8) {
  //
  // save the plain encoding
  //
  ofstream
    ofs(fn.data());
  if (!ofs) {
    cout << "Couldn't open the output file" << endl;
    exit(0);
  }
  //
  // save the lengths of the arrays
  //
  char_vector
    lengths;
  ostream_iterator<char>
    char_str(ofs);
  assert(v1.size() < 65536);
  assert(v2.size() < 65536);
  assert(v3.size() < 65536);
  assert(v4.size() < 65536);
  assert(v5.size() < 65536);
  assert(v6.size() < 65536);
  assert(v7.size() < 65536);
  assert(v8.size() < 65536);
  lengths.push_back((unsigned char)(v1.size() % 256));
  lengths.push_back((unsigned char)(v1.size() / 256));
  lengths.push_back((unsigned char)(v2.size() % 256));
  lengths.push_back((unsigned char)(v2.size() / 256));
  lengths.push_back((unsigned char)(v3.size() % 256));
  lengths.push_back((unsigned char)(v3.size() / 256));
  lengths.push_back((unsigned char)(v4.size() % 256));
  lengths.push_back((unsigned char)(v4.size() / 256));
  lengths.push_back((unsigned char)(v5.size() % 256));
  lengths.push_back((unsigned char)(v5.size() / 256));
  lengths.push_back((unsigned char)(v6.size() % 256));
  lengths.push_back((unsigned char)(v6.size() / 256));
  lengths.push_back((unsigned char)(v7.size() % 256));
  lengths.push_back((unsigned char)(v7.size() / 256));
  lengths.push_back((unsigned char)(v8.size() % 256));
  lengths.push_back((unsigned char)(v8.size() / 256));
  copy(lengths.begin(), lengths.end(), char_str);
  //
  // save the data of the arrays
  //
  copy(v1.begin(), v1.end(), char_str);
  copy(v2.begin(), v2.end(), char_str);
  copy(v3.begin(), v3.end(), char_str);
  copy(v4.begin(), v4.end(), char_str);
  copy(v5.begin(), v5.end(), char_str);
  copy(v6.begin(), v6.end(), char_str);
  copy(v7.begin(), v7.end(), char_str);
  copy(v8.begin(), v8.end(), char_str);
  ofs.close();
}


char_vector stat_encoder(char_vector orig) {
  //
  // Statistical Encoders: try different combination and choose
  // the best!
  //
  ZlibEngine
    eng;

  unsigned int
    ind,
    j,
    mini;

  char_vector
    rle  = rle_encoder(orig);

  //char_vector
  //  mtf  = mtf_encoder(orig);

  vector<char_vector>
    encodings;

  encodings.push_back(orig);
  encodings.push_back(eng.huffman_encoder(orig));
  encodings.push_back(arith_encoder(orig));
  encodings.push_back(eng.encoder(orig, 9)); // ZLIB BEST
  encodings.push_back(rle);
  encodings.push_back(eng.huffman_encoder(rle));
  encodings.push_back(arith_encoder(rle));
  encodings.push_back(eng.encoder(rle, 9)); // RLE + ZLIB BEST
  //encodings.push_back(eng.huffman_encoder(mtf));
  //encodings.push_back(arith_encoder(mtf));
  //encodings.push_back(eng.encoder(mtf, 9)); // MTF + ZLIB BEST

  mini = orig.size();
  ind = 0;
  j = 1;

  for (vector<char_vector>::iterator i = encodings.begin()+1; i !=encodings.end(); ++i, ++j) {
    if (i->size() < mini) {
      mini = i->size();
      ind = j;
    }
  }

# ifdef PRINTENCODER
  switch (ind) {
  case 0: cout << "Nothing";
    break;
  case 1: cout << "Huffman";
    break;
  case 2: cout << "Arithmetic";
    break;
  case 3: cout << "ZLib best";
    break;
  case 4: cout << "RLE";
    break;
  case 5: cout << "RLE + Huffman";
    break;
  case 6: cout << "RLE + Arithmetic";
    break;
  case 7: cout << "RLE + Zlib Best";
    break;
    //case 8: cout << "MTF + Huffman";
    //break;
    //case 9: cout << "MTF + Arithmetic";
    //break;
    //case 10: cout << "MTF + Zlib Best";
    //break;
  }
# endif
  return encodings[ind];
}
  
#ifdef OFF3
inline int final_size(Text & txt) { 
  //
  // compute the final size of OFF3 using statistical encoders
  //

  //
  // transform the leftover with the new representation
  //
  for (text_vector_iterator i = txt.text.begin(); i != txt.text.end(); ++i)
#   ifdef MAX_256
    ttext.push_back((unsigned short)(i->first & 0xFF));
#   else
    if (i->second == POINTER) {
      if ((i+1)->second != POINTER)
	cout << "Warning: no two consecutive pointers at " << i - txt.text.begin() << endl;
      ttext.push_back((unsigned char)(i->first & 0xFF));
      ttext_h.push_back((unsigned char)((i+1)->first & 0xFF));
      ++i;
    } else {
      ttext.push_back((unsigned char)(i->first & 0xFF));
      ttext_h.push_back((unsigned char)0);
    }
#   endif
  //
  // sort the lengths
  //
  sort(sublen_vec.begin(), sublen_vec.end());
  unsigned char
    last = *(sublen_vec.begin());
  delta.push_back(last);
  for (char_vector_iterator i = sublen_vec.begin(); i != sublen_vec.end(); ++i) {
      delta.push_back((unsigned char)(*i - last));
      last = *i;
  }
  cout << "Encoders (text,ddict,delta) : ";
  int
    h1 = stat_encoder(ttext).size(),
#   ifdef MAX_256
    h2 = 0,
#   else
    h2 = stat_encoder(ttext_h).size(),
#   endif
    h3 = stat_encoder(ddict).size(),
    h4 = stat_encoder(delta).size();
  cout << endl;
  //
  // undo
  //
  ttext.erase(ttext.begin(),ttext.end());
# ifndef MAX_256
  ttext_h.erase(ttext_h.begin(),ttext_h.end());
# endif
  delta.erase(delta.begin(),delta.end());
  cout << "STAT: text_l = " << h1 << " text_h = " << h2 << " dict " << h3 << " delta " << h4 << endl;
  return h1 + h2 + h3 + h4;
}

void finalize(string fn) {
  //
  // Statistical encoding of arrays
  //
  cout << "=======================================================" << endl;

  cout << "text_low = [";
  print_vec(ttext);
  cout << "] (size " << ttext.size() << ") compressed using ";
  char_vector
    i2 = stat_encoder(ttext);
  cout << " to " << i2.size() << " bytes (avg " << B * (double)i2.size()/(double)ttext.size() << " bps)" << endl << endl;

# ifndef MAX_256
  cout << "text_high = [";
  print_vec(ttext_h);
  cout << "] (size " << ttext_h.size() << ") compressed using ";
  char_vector
    i3 = stat_encoder(ttext_h);
  cout << " to " << i3.size() << " bytes (avg " << B * (double)i3.size()/(double)ttext_h.size() << " bps)" << endl << endl;
# endif

  /*
  int
    sum = 0;
  double
    prob,
    avg;
  for (vector<int>::iterator i = occurr_vec.begin(); i != occurr_vec.end(); ++i)
    sum += *i;
  cout << "sum=" << sum << endl;
  for (vector<int>::iterator i = occurr_vec.begin(); i != occurr_vec.end(); ++i) {
    prob = (double)*i/(double)sum;
    cout << i - occurr_vec.begin() << " f=" << *i << " p=" << prob << " len=" << -log(prob)/log(2) << endl;
    avg += prob * -log(prob)/log(2);
  }
  cout << "Pointers coded with average of " << avg << " bits per symbol ";
  cout << "for total " << avg * (double)ttext.size() / 8.0 << " bytes" << endl << endl; */

  cout << "dict = [";
  print_vec_char(ddict);
  cout << "] (size " << ddict.size() << ") compressed using ";
  char_vector
    i4 = stat_encoder(ddict);
  cout << " to " << i4.size() << " bytes (avg " << B * (double)i4.size()/(double)ddict.size() << " bps)" << endl << endl;

  /* cout << "sublen = [";
  print_vec(sublen_vec);
  cout << "] (size " << sublen_vec.size() << ") compressed using ";
  char_vector
    i6 = stat_encoder(sublen_vec);
  cout << " to " << i6.size() << " bytes (avg " << B * (double)i6.size()/(double)sublen_vec.size() << " bps)" << endl << endl; */

  cout << "sublen delta sorted = [";
  print_vec(delta);
  cout << "] (size " << delta.size() << ") compressed using ";
  char_vector
    i7 = stat_encoder(delta);
  cout << " to " << i7.size() << " bytes (avg " << B * (double)i7.size()/(double)delta.size() << " bps)" << endl << endl;

  cout << "=======================================================" << endl;
  //cout << "Final size (bytes) = " << i0.size() + i1.size() + i2.size() + i3.size() + i4.size() + i5.size() + i6.size() << endl;
  //cout << "Final size (bytes) = " << i0.size() + i1.size() + i2.size() + i3.size() + i4.size() + i6.size() << endl;
# ifdef MAX_256
  cout << "Final size delta (bytes) = " << i2.size() + i4.size() + i7.size() << endl;
# else
  cout << "Final size delta (bytes) = " << i2.size() + i3.size() + i4.size() + i7.size() << endl;
# endif
  end_time2();

  /*
  string
    fn1 = fn + ".lo";
  //,
  //fn2 = fn + ".hi",
  //fn3 = fn + ".di";
  ofstream
    ofs1(fn1.c_str());
  //,
  //ofs2(fn2.c_str()),
  //ofs3(fn3.c_str());
  //if ((!ofs1)||(!ofs2)||(!ofs3)) {
  if (!ofs1) {
    cout << "Couldn't open the outputs file" << endl;
    exit(0);
  }
  ostream_iterator<char>
    char_str1(ofs1);
  //,
  //char_str2(ofs2),
  //char_str3(ofs3);
  copy(ttext.begin(), ttext.end(), char_str1);
  //copy(ttext_h.begin(), ttext_h.end(), char_str2);
  //copy(ddict.begin(), ddict.end(), char_str3);
  ofs1.close();
  //ofs2.close();
  //ofs3.close();
  */
}

#else

inline int final_size() { 
  //
  // compute the final size of OFF1, the eight arrays using statistical encoders
  //
  int
    h1 = stat_encoder(ttext).size(),
    h2 = stat_encoder(sublen_vec).size(),
    h3 = stat_encoder(dict_vec).size(),
    h4 = stat_encoder(occurr_vec).size(),
    h5 = stat_encoder(abspol_vec).size() + stat_encoder(abspoh_vec).size() + stat_encoder(relpol_vec).size() + stat_encoder(relpoh_vec).size(); 
  hTEXT= B * (double)h1 / (double)ttext.size(); 
  hLEN = B * (double)h2 / (double)sublen_vec.size(); 
  hTXT = B * (double)h3 / (double)dict_vec.size(); 
  hOCC = B * (double)h4 / (double)occurr_vec.size(); 
  hPOS = 2.0 * B * (double)h5 / (double)(abspol_vec.size() + abspoh_vec.size() + relpol_vec.size() + relpoh_vec.size()); 
  return h1 + h2 + h3 + h4 + h5; 
}


void finalize(string fn) {
  //
  // Statistical encoding of arrays
  //
  cout << "=======================================================" << endl;

  cout << "text = [";
  print_vec_char(ttext);
  cout << "] (size " << ttext.size() << ") compressed using ";
  char_vector
    i0 = stat_encoder(ttext);
  cout << " to " << i0.size() << " bytes (avg " << B * (double)i0.size()/(double)ttext.size() << " bps)" << endl << endl;

  cout << "sublen = [";
  print_vec (sublen_vec);
  cout << "] (size " << sublen_vec.size() << ") compressed using ";
  char_vector
    i1 = stat_encoder(sublen_vec);
  cout << " to " << i1.size() << " bytes (avg " << B * (double)i1.size()/(double)sublen_vec.size() << " bps)" << endl << endl;

  cout << "dict = [";
  print_vec_char(dict_vec);
  cout << "] (size " << dict_vec.size() << ") compressed using ";
  char_vector
    i2 = stat_encoder(dict_vec);
  cout << " to " << i2.size() << " bytes (avg " << B * (double)i2.size()/(double)dict_vec.size() << " bps)" << endl << endl;

  cout << "occurr = [";
  print_vec (occurr_vec);
  cout << "] (size " << occurr_vec.size() << ") compressed using ";
  char_vector
    i3 = stat_encoder(occurr_vec);
  cout << " to " << i3.size() << " bytes (avg " << B * (double)i3.size()/(double)occurr_vec.size() << " bps)" << endl << endl;

  cout << "abspol = [";
  print_vec (abspol_vec);
  cout << "] (size " << abspol_vec.size() << ") compressed using ";
  char_vector
    i4_1 = stat_encoder(abspol_vec);
  cout << " to " << i4_1.size() << " bytes (avg " << B * (double)i4_1.size()/(double)abspol_vec.size() << " bps)" << endl << endl;

  cout << "abspoh = [";
  print_vec (abspoh_vec);
  cout << "] (size " << abspoh_vec.size() << ") compressed using ";
  char_vector
    i4_2 = stat_encoder(abspoh_vec);
  cout << " to " << i4_2.size() << " bytes (avg " << B * (double)i4_2.size()/(double)abspoh_vec.size() << " bps)" << endl << endl;

  cout << "relpol = [";
  print_vec (relpol_vec);
  cout << "] (size " << relpol_vec.size() << ") compressed using ";
  char_vector
    i5_1 = stat_encoder(relpol_vec);
  cout << " to " << i5_1.size() << " bytes (avg " << B * (double)i5_1.size()/(double)relpol_vec.size() << " bps)" << endl << endl;

  cout << "relpoh = [";
  print_vec (relpoh_vec);
  cout << "] (size " << relpoh_vec.size() << ") compressed using ";
  char_vector
    i5_2 = stat_encoder(relpoh_vec);
  cout << " to " << i5_2.size() << " bytes (avg " << B * (double)i5_2.size()/(double)relpoh_vec.size() << " bps)" << endl << endl;

  cout << "=======================================================" << endl;

  cout << "Final size (bytes) = " << 3 + i0.size() + i1.size() + i2.size() + i3.size() + i4_1.size() + i4_2.size() + i5_1.size() + i5_2.size() << endl;

  end_time2();

# ifdef SAVEENCODING
  string
    fn1 = file_name+".off";
  save_encoding (fn1, txt.text, sublen_vec, dict_vec, occurr_vec,
		 abspol_vec, abspoh_vec, relpol_vec, relpoh_vec, prefix_vec);
# endif

# ifdef SAVECOMPRESSEDENCODING
  string
    fn2 = file_name+".offc";
  save_encoding (fn2, i0, i1, i2, i3, i4_1, i4_2, i5_1, i5_2); // prefix is dummy
# endif
}
#endif


main(int argc, char* argv[]) {
  //
  // main
  //
  string
    file_name = ".in";
  int
    reco_no = 0, // # of tree constructions
    itera = 0,   // # of substitutions
    subst = 1;

  if (argc == 2)
     file_name = argv[1];

  ifstream
    ifs(file_name.data()); // ios::ate

  if (!ifs) {
    cout << "Couldn't open the input file" << endl;
    exit(0);
  }

# ifdef FINAL_ENCODING
  string
    log_name = file_name + ".log";

  ofstream
    logfs(log_name.data());
  if (!logfs) {
    cout << "Couldn't open the output file for log" << endl;
    exit(0);
  }
# endif

# ifdef VERBOSE
  cout << "Reading file" << endl;
# endif
  //
  // open and read the file into text
  //
  for (unsigned char c;;) {
    ifs.get(c);
    if (ifs)
      ttext.push_back(c);
    else
      break;
  }

# ifdef OFF3
  //
  // compute the alphabeth size
  //
  bool
    f[MAXCHAR+2];
  int
    alph_size = 0;

  for (int i = 0; i < MAXCHAR + 1; ++i)
    f[(unsigned char)(i & 0xFF)] = false;

  for (char_vector_iterator i = ttext.begin(); i != ttext.end(); ++i)
    f[*i] = true;

  for (int i = 0; i < MAXCHAR + 1; ++i)
    if (f[(unsigned char)(i & 0xFF)]) {
      cout << "char : " << (unsigned char)(i & 0xFF) << endl;
      alph_size++;
    }
# endif

  //
  // some welcome messages
  //
  cout << "Starting off-line compression on: " << file_name << endl;
  cout << "Threshold        = " << THRE << endl;
  cout << "Queue length     = " << QUEUE_LEN << endl;
  cout << "Min length match = " << MIN_LENGTH << endl;
  cout << "Max length match = " << MAX_LENGTH << endl;
  cout << "Min occurrences  = " << MIN_OCC << endl;
# ifdef MAX_256
  cout << "Max iterations   = 256" << endl;
# else
  cout << "Max iterations   = unlimited" << endl;
# endif
# ifdef OFF3
  cout << "Encoder          = Off-line 3" << endl;
# else
  cout << "Encoder          = Off-line 1" << endl;
# endif
# ifdef AUGMENTED
  cout << "Augmented suffix tree" << endl;
# else
  cout << "Standard suffix tree" << endl;
# endif

  //
  // time computation
  //
  start_time2();

  //
  // SUPERMAIN LOOP
  //
  //for (int step = 0; step < 2; step++) {  
  //cout << "STEP " << step << endl;
    Stat
      s;
    stat_queue
      sq(QUEUE_LEN);
    Text
      txt(ttext);
    Tree
      *aTree = NULL;

    int
      h = stat_encoder(ttext).size();
    hTEXT = B * (double)h / (double)ttext.size();
    hTXT  = 0.0;
    hLEN  = 0.0;
    hOCC  = 0.0;
    hPOS  = 0.0;

    ttext.erase(ttext.begin(), ttext.end());

    cout << " compression (bytes) = " << h << " (avg " << hTEXT << " bps)" << endl;
    
#   ifdef OFF3
    cout << "Alphabeth size   = " << alph_size << " log_2(size)=" << log(alph_size)/log(2) << endl;
#   endif

#   ifdef FINAL_ENCODING
    int
      new_h = h;
    //logfs << "itera" << (char)(0x09) << text_len << (char)(0x09) << dicting.size() << (char)(0x09) << p.size() << (char)(0x09) << s.get_freq() << (char)(0x09) << s.get_gain() << (char)(0x09) << new_h << endl;
    logfs << "i" << (char)(0x09) << "hTEXT" << (char)(0x09) << "hLEN" << (char)(0x09) << "hTXT" << (char)(0x09) << "hOCC" << (char)(0x09) << "hPOS" << (char)(0x09) << "final" << endl;
#   endif

#   ifdef OFF3
    //
    // filter text 
    //
    int
      freqchar[MAXCHAR+2];
    bool
      used[MAXCHAR+2];
    // 
    // build the alphabet
    //
    for (int i = 0; i < MAXCHAR + 1; ++i)
      freqchar[(unsigned char)(i & 0xFF)] = 0;
    
    for (text_vector_iterator i = txt.text.begin(); i != txt.text.end(); ++i)
      if (i->second == LITERAL)
	freqchar[(unsigned char)(i->first & 0xFF)]++;
    
    //
    // assign each symbol in the alphabet a new pointer
    //
    for (int i = 0; i < MAXCHAR + 1; ++i)
      used[(unsigned char)(i & 0xFF)] = (freqchar[(unsigned char)(i & 0xFF)] > 0);
#   endif
    //
    // MAIN LOOP
    //
    while (1) { 
      s.init();

#     ifdef PRINTTEXT
      cout << "---- Text(" << txt.text.size() << ") ----" << endl;
#     ifdef OFF3
      print_text_vec(txt.text);
#     else
      print_vec_char(txt.text);
#     endif
      cout << endl << endl;
#     endif

      //
      // check the queue of strings
      //
      if (sq.empty()) {
	if (aTree != NULL)
	  aTree->Tree::~Tree();

	if (!subst) { // nothing has been replaced: go out!
	  cout << "Empty queue!" << endl;
	  break;
	}

	//
	// create the augmented suffix tree
	//
#       ifdef VERBOSE
	cout << "Creating tree" << endl;
#       endif

	reco_no++;
#       ifdef NO_RECURSIVE_PTR
	text_vector_iterator
	  tt_i, tt_j;
	for (tt_i = txt.text.begin(); tt_i != txt.text.end(); ++tt_i) {
	  for (tt_j = tt_i; tt_j != txt.text.end(); ++tt_j)
	    if (tt_j->second == POINTER)
	      break;
	  if (tt_j != tt_i)
	    break;
	}
	// cout << tt_i - txt.text.begin() << "->" <<  tt_j - txt.text.begin() << endl;
	if (tt_i == txt.text.end()) { // all pointers!
	  cout << "No leftover" << endl;
	  break;
	}
#       ifdef PRUNED
        aTree = new Tree(tt_i, min(tt_j, tt_i + MAX_LENGTH));
#       else
	aTree = new Tree(tt_i, tt_j);
#       endif
#       ifdef AUGMENTED
	aTree->build_augmented_suffix_tree(tt_i + 1, txt.text);
#       else
	aTree->build_suffix_tree(txt.text);
#       endif

#       else // RECURSIVE PTRS ALLOWED

#       ifdef PRUNED
	aTree = new Tree(txt.text.begin(), min(txt.text.end(), txt.text.begin() + MAX_LENGTH));
#       else
        aTree = new Tree(txt.text.begin(), txt.text.end());
#       endif
#       ifdef AUGMENTED
        aTree->build_augmented_suffix_tree(txt.text.begin() + 1, txt.text);
#       else
        aTree->build_suffix_tree(txt.text);
#       endif
#       endif
	
	//
	// compute statistics
	//
	//aTree->clear_terminators();
#       ifdef PRINTTREE
	cout << "---- Tree (Non Overlapped Occurrences) ----" << endl;
	aTree->print(txt.text, 1);
	cout << endl;
#       endif
	aTree->compute_gain(sq, txt.text, 0);
	subst = 0;
	if (sq.empty()) {
	  cout << "Empty queue." << endl;
	  break;
	}
#       ifdef PRINTQUEUE
	sq.print_queue();
#       endif
      }
      
      //
      // get the next candidate
      // 
      if (!sq.empty()) {
	s = sq.top();
	sq.pop();
      }
      
      //
      // is it worthwhile substituting? 
      //
      if (s.get_gain() > THRE) {
	text_vector
	  substring = s.get_string();
	
	int
	  sub_len = substring.size(),
	  text_len = txt.text.size();
	
#       ifdef VERBOSE
	cout << "Searching and compacting" << endl;
#       endif
	
#       ifdef OFF3 // ---------------------- OFF3 ----------------------------
	text_vector
	  ptr;
	
	/* unsigned int
	   position = std::search(dict_vec.begin(), dict_vec.end(), substring.begin(), substring.end()) - dict_vec.begin();
	   
	   ptr.push_back(symbol((unsigned char)(position % 256), POINTER));
	   ptr.push_back(symbol((unsigned char)(position / 256), POINTER)); */
	//
	// search for the first non-used pointer
	//
	while (used[(unsigned char)itera]) {
	  cout << "(" << itera << ") Appending to dictionary \"" << (unsigned char)itera << "\"" << endl;
	  ddict.push_back((unsigned char)itera);
	  sublen_vec.push_back((unsigned char)1);
	  used[(unsigned char)itera] = false;
	  ++itera;
	}
	ptr.push_back(symbol((unsigned char)(itera % 256), POINTER));
#       ifndef MAX_256
	ptr.push_back(symbol((unsigned char)(itera / 256), POINTER));
#       endif
	
	vector<int>
	  p = txt.substitute(substring, ptr);
	
	if (p.size() >= MIN_OCC) {
	  // cout << "Position " << position << endl;
	  // if (position == dict_vec.size())
	  for (text_vector_iterator ss_i = substring.begin(); ss_i != substring.end(); ++ss_i)
	    ddict.push_back(ss_i->first);
	  // else cout << " Dictionary not updated!" << endl;
	  int
	    sl = sub_len - MIN_LENGTH;
	  while (sl >= MAXCHAR) {
	    sublen_vec.push_back((unsigned char)(MAXCHAR));
	    sl -= MAXCHAR;
	  }
	  sublen_vec.push_back((unsigned char)(sl));
	  //occurr_vec.push_back(p.size() - 1);
	  cout << "(" << itera << ") Text_len: " << text_len << endl;
	  cout << "Sub_string: \"";
	  print_text_vec(s.get_string());
	  cout << "\"[" << sub_len;
	  cout << "], n_occ: " << p.size() << " gain: " << s.get_gain();
	  cout << " freq: " << s.get_freq() << endl;
#         ifdef PRINTTEXT
	  cout << "Dictionary: \"";
	  print_text_vec(dict_vec);
	  cout << "\"[" << dict_vec.size() << "]" << endl;
	  cout << "Len: \"";
	  print_vec(sublen_vec);
	  cout << "\"[" << sublen_vec.size() << "]" << endl;
#         endif
	  cout << endl;
	  //
	  // compact text
	  //
	  txt.defragment(); 
	  itera++;
	  subst++;

#         ifdef RECOMPUTE_ENCODING 
	  //
	  // recompute final encoding at each step
	  //
	  int
	    new_h = final_size(txt);
	  cout << "(" << itera-1 << ") Final size " << new_h << endl << ":::::::::::::::::::::::::::" << endl;
#         endif

#         ifdef MAX_256
	  if (itera > MAXCHAR) {
	    cout << "Max Iterations reached." << endl;
	    break;
	  }
#         endif
	}
	
#       else // ---------------------- OFF1 ----------------------------
	
	vector<int>
	  p = txt.erase(substring);
	
	if (p.size() >= MIN_OCC) { // at least MIN_OCC occurrence!
	  subst++;
	  //
	  // save the first (absolute) occurrence
	  // MAXSHORT means read another doublebyte
	  //
	  itera++;
	  int
	    pos = *(p.begin());
	  while (pos >= MAXSHORT) {
	    abspol_vec.push_back((unsigned char)(MAXCHAR));
	    abspoh_vec.push_back((unsigned char)(MAXCHAR));
	    pos -= MAXSHORT;
	  }
	  abspol_vec.push_back((unsigned char)(pos % 256));
	  abspoh_vec.push_back((unsigned char)(pos / 256));
	  //
	  // search for all the other occurrences
	  //
	  int
	    last = *(p.begin()); // last relative occurrence
	  for (vector<int>::iterator pi = p.begin() + 1; pi != p.end(); ++pi) {
	    //
	    // save the relative occurrence
	    // MAXSHORT means read another doublebyte
	    // 
	    last = *pi - last - sub_len;
	    assert (last >= 0);
	    while (last >= MAXSHORT) {
	      relpol_vec.push_back((unsigned char)(MAXCHAR));
	      relpoh_vec.push_back((unsigned char)(MAXCHAR));
	      last -= MAXSHORT;
	    }
	    relpol_vec.push_back((unsigned char)(last % 256));
	    relpoh_vec.push_back((unsigned char)(last / 256));
	    last = *pi;
	  }
	  
	  int
	    occ = p.size() - MIN_OCC;
	  
	  while (occ >= MAXCHAR) {
	    //
	    // save occurrences - MIN_OCC, MAXCHAR means read another byte
	    //
	    occurr_vec.push_back((unsigned char)(MAXCHAR));
	    //
	    // I save (sub_len-MIN_LENGTH), but MAXCHAR means read another byte
	    //
	    int
	      sl = sub_len - MIN_LENGTH;
	    while (sl >= MAXCHAR) {
	      sublen_vec.push_back((unsigned char)(MAXCHAR));
	      sl -= MAXCHAR;
	    }
	    sublen_vec.push_back((unsigned char)(sl));
	    for (text_vector_iterator ss_i = substring.begin(); ss_i != substring.end(); ++ss_i) 
	      dict_vec.push_back((unsigned char)(*ss_i));
	    
	    occ -= MAXCHAR;
	  }
	  //
	  // save the remaining number of occurrence,
	  // the substring length and the substring
	  //
	  occurr_vec.push_back((unsigned char)(occ));
	  int
	    sl = sub_len - MIN_LENGTH;
	  while (sl >= MAXCHAR) {
	    sublen_vec.push_back((unsigned char)(MAXCHAR));
	    sl -= MAXCHAR;
	  }
	  sublen_vec.push_back((unsigned char)(sl));
	  for (text_vector_iterator ss_i = substring.begin(); ss_i != substring.end(); ++ss_i) 
	    dict_vec.push_back((unsigned char)(*ss_i));
	  
	  cout << "Text_len: " << text_len << endl;
	  cout << "Sub_string: \"";
	  print_text_vec(s.get_string());
	  cout << "\"[" << sub_len;
	  cout << "], n_occ: " << p.size() << " gain: " << s.get_gain();
	  cout << " freq: " << s.get_freq() << endl << endl; 
	  
	  //
	  // compact text
	  //
	  txt.defragment();
	  
#         ifdef FINAL_ENCODING
	  //
	  // recompute final encoding
	  //
	  ttext = txt.to_char_vector();
	  new_h = final_size();
	  cout << "Final Size " << new_h << endl << endl;
	  
	  //logfs << itera << (char)(0x09) << text_len << (char)(0x09) << substring.size() << (char)(0x09) << p.size() << (char)(0x09) << s.get_freq() << (char)(0x09) << s.get_gain() << (char)(0x09) << new_h << endl;
	  logfs << itera << (char)(0x09) << hTEXT << (char)(0x09) << hLEN << (char)(0x09) << hTXT << (char)(0x09) << hOCC << (char)(0x09) << hPOS << (char)(0x09) << new_h << endl;
#         endif // FINAL_ENCODING
	}
#       endif // OFF1-3
      }
    };
    
#   ifdef OFF3
    for (int i = 0; i < MAXCHAR + 1; ++i)
      if (used[(unsigned char)i]) {
	cout << "(" << itera << ") Appending to dictionary \"" << (unsigned char)i << "\"" << endl;
	ddict.push_back((unsigned char)i);
	sublen_vec.push_back((unsigned char)1);
	for (text_vector_iterator it = txt.text.begin(); it != txt.text.end(); ++it)
	  if ((it->first == (unsigned char)i)&&(it->second == LITERAL))
	    it->first = (unsigned char)(itera & 0xFF);
	used[(unsigned char)i] = false;
	++itera;
      }
    //
    // transform the leftover with the new representation
    //
    for (text_vector_iterator i = txt.text.begin(); i != txt.text.end(); ++i)
#     ifdef MAX_256
      ttext.push_back((unsigned short)(i->first & 0xFF));
#     else
      if (i->second == POINTER) {
        if ((i+1)->second != POINTER)
	  cout << "Warning: no two consecutive pointers at " << i - txt.text.begin() << endl;
	ttext.push_back((unsigned char)(i->first & 0xFF));
	ttext_h.push_back((unsigned char)((i+1)->first & 0xFF));
	++i;
      } else {
	ttext.push_back((unsigned char)(i->first & 0xFF));
	ttext_h.push_back((unsigned char)0);
      }
#     endif
    //
    // sort the lengths
    //
    sort(sublen_vec.begin(), sublen_vec.end());
    unsigned char
      last = *(sublen_vec.begin());
    delta.push_back(last);
    for (char_vector_iterator i = sublen_vec.begin(); i != sublen_vec.end(); ++i) {
      delta.push_back((unsigned char)(*i - last));
      last = *i;
    }
#   else
    ttext = txt.text;
#   endif
    //}
    finalize(file_name);
    cout << "Iteractions = " << itera << endl;
    cout << "Number of tree constructions = " << reco_no << endl;
    cout << "Iteractions/Number of tree constructions = " << (double)itera/(double)reco_no << endl;  
# ifdef FINAL_ENCODING
    logfs.close();
# endif
}
