//
// Statistics Construction
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v2.0 10.28.98 $
//

#include <assert.h>
#include <fstream.h>
#include "st.h"
#include "dyna.h"
//#include "pm.h"
#include "zlibengn.h"
#include "arith.h"

#define AUGMENTED
// #define SAVEENCODING
// #define SAVECOMPRESSEDENCODING
// #define DEBUGFINALTREE
// #define PRINTTEXT
// #define PRINTTREE
// #define PRINTQUEUE
#define PRINTENCODER
// #define DEBUGSTAT
// #define DEBUG
// #define VERBOSE // print all the step performed

double
  hTEXT,
  hTXT,
  hOCC,
  hLEN,
  hPOS;
base2
  loga;

char_vector
  ttext,
  ttext_b,
  ttext_l,
  ttext_h,
  sublen_vec;

extern void start_time2 (void);
extern void end_time2 (void);

void print_vec (char_vector& x) {
  //
  // print a vector<char> in numbers
  //
  for (char_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << " " << (unsigned short)(*i & 0xFF);
}



void print_text_vec (text_vector& x) {
  //
  // print a vector<char> in chars
  //
  for (text_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << *i;
}



void print_vec_char (char_vector& x) {
  //
  // print a vector<char> in chars
  //
  for (char_vector_iterator i = x.begin(); i != x.end(); ++i)
    cout << *i;
}



void save_encoding(string fn, char_vector& v1, char_vector& v2,
		   char_vector& v3, char_vector& v4, char_vector& v5,
		   char_vector& v6, char_vector& v7, char_vector& v8) {
  //
  // save the plain encoding
  //
  ofstream
    ofs(fn.data());
  if (!ofs) {
    cout << "Couldn't open the output file" << endl;
    exit(0);
  }
  //
  // save the lengths of the arrays
  //
  char_vector
    lengths;
  ostream_iterator<char>
    char_str(ofs);
  assert(v1.size() < 65536);
  assert(v2.size() < 65536);
  assert(v3.size() < 65536);
  assert(v4.size() < 65536);
  assert(v5.size() < 65536);
  assert(v6.size() < 65536);
  assert(v7.size() < 65536);
  assert(v8.size() < 65536);
  lengths.push_back((unsigned char)(v1.size() % 256));
  lengths.push_back((unsigned char)(v1.size() / 256));
  lengths.push_back((unsigned char)(v2.size() % 256));
  lengths.push_back((unsigned char)(v2.size() / 256));
  lengths.push_back((unsigned char)(v3.size() % 256));
  lengths.push_back((unsigned char)(v3.size() / 256));
  lengths.push_back((unsigned char)(v4.size() % 256));
  lengths.push_back((unsigned char)(v4.size() / 256));
  lengths.push_back((unsigned char)(v5.size() % 256));
  lengths.push_back((unsigned char)(v5.size() / 256));
  lengths.push_back((unsigned char)(v6.size() % 256));
  lengths.push_back((unsigned char)(v6.size() / 256));
  lengths.push_back((unsigned char)(v7.size() % 256));
  lengths.push_back((unsigned char)(v7.size() / 256));
  lengths.push_back((unsigned char)(v8.size() % 256));
  lengths.push_back((unsigned char)(v8.size() / 256));
  copy(lengths.begin(), lengths.end(), char_str);
  //
  // save the data of the arrays
  //
  copy(v1.begin(), v1.end(), char_str);
  copy(v2.begin(), v2.end(), char_str);
  copy(v3.begin(), v3.end(), char_str);
  copy(v4.begin(), v4.end(), char_str);
  copy(v5.begin(), v5.end(), char_str);
  copy(v6.begin(), v6.end(), char_str);
  copy(v7.begin(), v7.end(), char_str);
  copy(v8.begin(), v8.end(), char_str);
  ofs.close();
}


char_vector stat_encoder(char_vector orig) {
  //
  // Statistical Encoders: try different combination and choose
  // the best!
  //
  ZlibEngine
    eng;

  unsigned int
    ind,
    j,
    mini;

  char_vector
    rle  = rle_encoder(orig);

  //char_vector
  //  mtf  = mtf_encoder(orig);

  vector<char_vector>
    encodings;

  encodings.push_back(orig);
  encodings.push_back(eng.huffman_encoder(orig));
  encodings.push_back(arith_encoder(orig));
  encodings.push_back(eng.encoder(orig, 9)); // ZLIB BEST
  encodings.push_back(rle);
  encodings.push_back(eng.huffman_encoder(rle));
  encodings.push_back(arith_encoder(rle));
  encodings.push_back(eng.encoder(rle, 9)); // RLE + ZLIB BEST
  //encodings.push_back(eng.huffman_encoder(mtf));
  //encodings.push_back(arith_encoder(mtf));
  //encodings.push_back(eng.encoder(mtf, 9)); // MTF + ZLIB BEST

  mini = orig.size();
  ind = 0;
  j = 1;

  for (vector<char_vector>::iterator i = encodings.begin()+1; i !=encodings.end(); ++i, ++j) {
    if (i->size() < mini) {
      mini = i->size();
      ind = j;
    }
  }

# ifdef PRINTENCODER
  switch (ind) {
  case 0: cout << "Nothing";
    break;
  case 1: cout << "Huffman";
    break;
  case 2: cout << "Arithmetic";
    break;
  case 3: cout << "ZLib best";
    break;
  case 4: cout << "RLE";
    break;
  case 5: cout << "RLE + Huffman";
    break;
  case 6: cout << "RLE + Arithmetic";
    break;
  case 7: cout << "RLE + Zlib Best";
    break;
    //case 8: cout << "MTF + Huffman";
    //break;
    //case 9: cout << "MTF + Arithmetic";
    //break;
    //case 10: cout << "MTF + Zlib Best";
    //break;
  }
# endif
  return encodings[ind];
}
  
/*
inline int final_size() { 
  //
  // compute the final size of OFF1, the eight arrays using statistical encoders
  //
  int
    h1 = stat_encoder(ttext).size(),
    h2 = stat_encoder(ttext.second).size(),
    h3 = stat_encoder(dict_vec.first).size(),
    h4 = stat_encoder(dict_vec.second).size(),
    h5 = stat_encoder(sublen_vec).size(); 
  return h1 + h2 + h3 + h4 + h5; 
}*/

void finalize(string fn) {
  //
  // Statistical encoding of arrays
  //
  cout << "=======================================================" << endl;

  cout << "text = [";
  print_vec_char(ttext);
  cout << "] (size " << ttext.size() << ") compressed using ";
  char_vector
    i0 = stat_encoder(ttext);
  cout << " to " << i0.size() << " bytes (avg " << B * (double)i0.size()/(double)ttext.size() << " bps)" << endl << endl;

  cout << "text_bit_vector = [";
  print_vec(ttext_b);
  cout << "] (size " << ttext_b.size() << ") compressed using ";
  char_vector
    i1 = stat_encoder(ttext_b);
  cout << " to " << i1.size() << " bytes (avg " << B * (double)i1.size()/(double)ttext_b.size() << " bps)" << endl << endl;

  cout << "text_low = [";
  print_vec(ttext_l);
  cout << "] (size " << ttext_l.size() << ") compressed using ";
  char_vector
    i2 = stat_encoder(ttext_l);
  cout << " to " << i2.size() << " bytes (avg " << B * (double)i2.size()/(double)ttext_l.size() << " bps)" << endl << endl;

  cout << "text_high = [";
  print_vec(ttext_h);
  cout << "] (size " << ttext_h.size() << ") compressed using ";
  char_vector
    i3 = stat_encoder(ttext_h);
  cout << " to " << i3.size() << " bytes (avg " << B * (double)i3.size()/(double)ttext_h.size() << " bps)" << endl << endl;

  cout << "sublen = [";
  print_vec(sublen_vec);
  cout << "] (size " << sublen_vec.size() << ") compressed using ";
  char_vector
    i4 = stat_encoder(sublen_vec);
  cout << " to " << i4.size() << " bytes (avg " << B * (double)i4.size()/(double)sublen_vec.size() << " bps)" << endl << endl; 

  cout << "=======================================================" << endl;
  cout << "Final size (bytes) = " << i0.size() + i1.size() + i2.size() + i3.size() + i4.size() << endl;
  end_time2();
}



main(int argc, char* argv[]) {
  //
  // main
  //
  string
    file_name = ".in";
  int
    reco_no = 0, // # of tree constructions
    itera = 0,   // # of substitutions
    subst = 1;

  if (argc == 2)
     file_name = argv[1];

  ifstream
    ifs(file_name.c_str()); // ios::ate

  if (!ifs) {
    cout << "Couldn't open the input file" << endl;
    exit(0);
  }

# ifdef VERBOSE
  cout << "Reading file" << endl;
# endif
  //
  // open and read the file into text
  //
  for (unsigned char c;;) {
    ifs.get(c);
    if (ifs)
      ttext.push_back(c);
    else
      break;
  }
  ttext.push_back(TERMINATOR);

  //
  // compute the alphabeth size
  //
  bool
    f[MAXCHAR+2];
  int
    alph_size = 0;

  for (int i = 0; i < MAXCHAR + 1; ++i)
    f[(unsigned char)(i & 0xFF)] = false;

  for (char_vector_iterator i = ttext.begin(); i != ttext.end(); ++i)
    f[*i] = true;

  for (int i = 0; i < MAXCHAR + 1; ++i)
    if (f[(unsigned char)(i & 0xFF)])
      alph_size++;

  //
  // some welcome messages
  //
  cout << "Starting off-line compression on: " << file_name << endl;
  cout << "Threshold        = " << THRE << endl;
  cout << "Queue length     = " << QUEUE_LEN << endl;
  cout << "Min length match = " << MIN_LENGTH << endl;
  cout << "Max length match = " << MAX_LENGTH << endl;
  cout << "Min occurrences  = " << MIN_OCC << endl;
  cout << "Encoder          = Off-line 2" << endl;

# ifdef AUGMENTED
  cout << "Augmented suffix tree" << endl;
# else
  cout << "Standard suffix tree" << endl;
# endif

  //
  // time computation
  //
  start_time2();

  //
  // SUPERMAIN LOOP
  //
  //for (int step = 0; step < 2; step++) {  
  //cout << "STEP " << step << endl;
    Stat
      s;
    stat_queue
      sq(QUEUE_LEN);
    Text
      txt(ttext);
    Tree
      *aTree = NULL;

    int
      h = stat_encoder(ttext).size();
    hTEXT = B * (double)h / (double)ttext.size();
    hTXT  = 0.0;
    hLEN  = 0.0;
    hOCC  = 0.0;
    hPOS  = 0.0;

    ttext.erase(ttext.begin(), ttext.end());

    cout << " compression (bytes) = " << h << " (avg " << hTEXT << " bps)" << endl;
    
    cout << "Alphabeth size   = " << alph_size << endl;

    //
    // MAIN LOOP
    //
    while (1) { 
      s.init();

#     ifdef PRINTTEXT
      cout << "---- Text(" << txt.text.size() << ") ----" << endl;
      print_text_vec(txt.text);
      cout << endl << endl;
#     endif

      //
      // check the queue of strings
      //
      if (sq.empty()) {
	if (aTree != NULL)
	  aTree->Tree::~Tree();

	if (!subst) { // nothing has been replaced: go out!
	  cout << "Empty queue!" << endl;
	  break;
	}

	//
	// create the augmented suffix tree
	//
#       ifdef VERBOSE
	cout << "Creating tree" << endl;
#       endif

	reco_no++;
#       ifdef NO_RECURSIVE_PTR
	text_vector_iterator
	  tt_i, tt_j;
	for (tt_i = txt.text.begin(); tt_i != txt.text.end(); ++tt_i) {
	  for (tt_j = tt_i; tt_j != txt.text.end(); ++tt_j)
	    if (tt_j->second == POINTER)
	      break;
	  if (tt_j != tt_i)
	    break;
	}
	// cout << tt_i - txt.text.begin() << "->" <<  tt_j - txt.text.begin() << endl;
	if (tt_i == txt.text.end()) { // all pointers!
	  cout << "No leftover" << endl;
	  break;
	}
#       ifdef PRUNED
        aTree = new Tree(tt_i, min(tt_j, tt_i + MAX_LENGTH));
#       else
	aTree = new Tree(tt_i, tt_j);
#       endif
#       ifdef AUGMENTED
	aTree->build_augmented_suffix_tree(tt_i + 1, txt.text);
#       else
	aTree->build_suffix_tree(txt.text);
#       endif

#       else // RECURSIVE PTRS ALLOWED

#       ifdef PRUNED
	aTree = new Tree(txt.text.begin(), min(txt.text.end(), txt.text.begin() + MAX_LENGTH));
#       else
        aTree = new Tree(txt.text.begin(), txt.text.end());
#       endif
#       ifdef AUGMENTED
        aTree->build_augmented_suffix_tree(txt.text.begin() + 1, txt.text);
#       else
        aTree->build_suffix_tree(txt.text);
#       endif
#       endif
	
	//
	// compute statistics
	//
#       ifdef PRINTTREE
	cout << "---- Tree (Non Overlapped Occurrences) ----" << endl;
	aTree->print(txt.text, 1);
	cout << endl;
#       endif
	aTree->compute_gain(sq, txt.text, 0);
	subst = 0;
	if (sq.empty()) {
	  cout << "Empty queue." << endl;
	  break;
	}
#       ifdef PRINTQUEUE
	sq.print_queue();
#       endif
      }
      
      //
      // get the next candidate
      // 
      if (!sq.empty()) {
	s = sq.top();
	sq.pop();
      }
      
      //
      // is it worthwhile substituting? 
      //
      if (s.get_gain() > THRE) {
	text_vector
	  substring = s.get_string();
	
	int
	  sub_len = substring.size(),
	  text_len = txt.text.size();
	
#       ifdef VERBOSE
	cout << "Searching and compacting" << endl;
#       endif
	
	vector<int>
	  p = txt.substitute2(substring);
	
	if (p.size() >= MIN_OCC) {
	  int
	    sl = sub_len - MIN_LENGTH;
	  while (sl >= MAXCHAR) {
	    sublen_vec.push_back((unsigned char)MAXCHAR);
	    sl -= MAXCHAR;
	  }
	  cout << "(" << itera << ") Text_len: " << text_len << endl;
	  cout << "Sub_string: \"";
	  print_text_vec(s.get_string());
	  cout << "\"[" << sub_len;
	  cout << "], n_occ: " << p.size() << " gain: " << s.get_gain();
	  cout << " freq: " << s.get_freq() << endl;
	  cout << endl;
	  //print_text_vec(txt.text);
	  //cout << endl;
	  //
	  // compact text
	  //
	  txt.defragment(); 
	  itera++;
	  subst++;
	  //if (itera > MAXCHAR) {
	  //  cout << "Max Iterations reached." << endl;
	  //  break;
	  //}
	}
      }
    };
    
    //
    // transform the leftover with the new representation
    //
    for (text_vector_iterator i = txt.text.begin(); i != txt.text.end(); ++i)
      if (i->second == POINTER) {
        if (((i+1)->second != POINTER) && ((i+2)->second != POINTER))
	  cout << "Warning: no three consecutive pointers at " << i - txt.text.begin() << endl;
	ttext_l.push_back((unsigned char)(i->first & 0xFF));
	ttext_b.push_back(i->second);
	ttext_h.push_back((unsigned char)((i+1)->first & 0xFF));
	sublen_vec.push_back((unsigned char)((i+2)->first & 0xFF));
	i += 2;
      } else {
	ttext.push_back((unsigned char)(i->first & 0xFF));
	ttext_b.push_back(i->second);
      }
    finalize(file_name);
    cout << "Iteractions = " << itera << endl;
    cout << "Number of tree constructions = " << reco_no << endl;
    cout << "Iteractions/Number of tree constructions = " << (double)itera/(double)reco_no << endl;  
}
