//
// Quick Pattern Search
// (Sunday CACM 1990, 33(8):132--142)
// using C++ STL generic function library
//
// Copyright (c) 1997, 1998, S. Lonardi, (stelo@cs.purdue.edu)
// May be used for research or personal use, but not for commercial
// use or as part of commercial software without written consent
// from the author.
//
// $ id = v1.2 1.1.98 $
//

#if !defined(_PM_H)
#define _PM_H

#define ASIZE 256
//#define DEBUG_QS

// ------------------------------------------------------------------------

template <class ForwardIterator1, class ForwardIterator2, class Distance1,
	  class Distance2>
ForwardIterator1 __QSsearch(ForwardIterator1 first1, ForwardIterator1 last1,
			  ForwardIterator2 first2, ForwardIterator2 last2,
			  Distance1*, Distance2*) {
    Distance1 d1 = 0;
    distance(first1, last1, d1);
    Distance2 d2 = 0;
    distance(first2, last2, d2);

    if (d1 < d2) return last1;

    int
      disp,
      table[ASIZE];

    for (int i = 0; i < ASIZE; i++)
      table[i] = d2 + 1;

    for (ForwardIterator2 current = first2; current != last2; ++current)
      table[(int)(*current)] = d2 - (current - first2);

#ifdef DEBUG_QS
    cout << "d1 = "<< d1 << endl;
    cout << "d2 = "<< d2 << endl;
    for (int i = 0; i < ASIZE; i++)
      cout << i << " " << table[i] << endl;
#endif

    ForwardIterator1 current1 = first1;
    ForwardIterator2 current2 = first2;

    while (current2 != last2)
      if (*current1++ != *current2++) {
	if (d1 < d2)
	  return last1;
	disp = table[(int)(*(first1 + d2))];
	d1 -= disp;
	first1  += disp; 
	current1 = first1;
	current2 = first2;
      }
    return (current2 == last2) ? first1 : last1;
#       ifdef DEBUG_QS
	cout << "mismatch at " << current1 - first1 << endl;
#       endif
#       ifdef DEBUG_QS
	cout << "shift by = " << disp << " char=" << *(first1 + d2) << endl;
#       endif
#       ifdef DEBUG_QS
	cout << "match at " << current1 - first1 << endl;
#       endif
}


template <class ForwardIterator1, class ForwardIterator2>
inline ForwardIterator1 QSsearch(ForwardIterator1 first1, ForwardIterator1 last1,
			       ForwardIterator2 first2, ForwardIterator2 last2)
{
    return __QSsearch(first1, last1, first2, last2,
		      distance_type(first1), distance_type(first2));
}

// ------------------------------------------------------------------------

template <class RandomAccessIterator, class Distance>
void compute_next(RandomAccessIterator pattern, 
                  RandomAccessIterator patternEnd,
                  vector<Distance>& next)
{
  Distance pattern_size = patternEnd - pattern, j = 0, t = -1;
  next.reserve(32);
  next.push_back(-1);
  while (j < pattern_size - 1) {
    while (t >= 0 && pattern[j] != pattern[t]) 
       t = next[t];
    ++j; ++t;
    if (pattern[j] == pattern[t]) 
      next.push_back(next[t]);
    else
      next.push_back(t);
  }
}


template <class ForwardIterator, class Distance>
void compute_next(ForwardIterator pattern, 
                  ForwardIterator patternEnd,
                  vector<Distance>& next, 
                  vector<ForwardIterator>& pattern_iterator)
{
  Distance t = -1;
  next.reserve(32);
  pattern_iterator.reserve(32);
  next.push_back(-1);
  pattern_iterator.push_back(pattern);
  for (;;) {
    ForwardIterator advance = pattern;
    ++advance;
    if (advance == patternEnd)
      break;
    while (t >= 0 && *pattern != *pattern_iterator[t]) 
       t = next[t];
    ++pattern; ++t;
    if (*pattern == *pattern_iterator[t]) 
      next.push_back(next[t]);
    else
      next.push_back(t);
    pattern_iterator.push_back(pattern);
  }
}


template <class ForwardIterator1, class ForwardIterator2, class Distance1, class Distance2>
ForwardIterator1 __search_L(ForwardIterator1 text, ForwardIterator1 textEnd,
                            ForwardIterator2 pattern, ForwardIterator2 patternEnd,
			    Distance1*, Distance2*)
{
  ForwardIterator1 advance, hold;
  ForwardIterator2 p, p1;
  Distance2 j, m;
  vector<Distance2> next;
  vector<ForwardIterator2> pattern_iterator;
  compute_next(pattern, patternEnd, next, pattern_iterator);
  m = next.size();
  if (next.size() == 1)
    return find(text, textEnd, *pattern);
  p1 = pattern; ++p1;
  while (text != textEnd) {
    while (*text != *pattern) 
      if (++text == textEnd)
        return textEnd;
    p = p1; j = 1;
    hold = text;
    if (++text == textEnd)
      return textEnd;
    while (*text == *p) {
      if (++p == patternEnd) 
        return hold;
      if (++text == textEnd)
        return textEnd;
      ++j;
    }
    for (;;) {
      j = next[j];
      if (j < 0) {
        ++text;
        break;
      }
      if (j == 0)
        break;
      p = pattern_iterator[j];
      while (*text == *p) {
        ++text; ++p; ++j;
        if (p == patternEnd) {
	  advance = hold;
          for (int i = m; --i >= 0;) 
            ++advance;
          while (advance != text) 
            ++advance, ++hold;
          return hold;
        }
        if (text == textEnd)
          return textEnd;
      }
    }
  }
  return textEnd;
}

template <class BidirectionalIterator1, class BidirectionalIterator2>
inline BidirectionalIterator1 Lsearch(BidirectionalIterator1 text,
				      BidirectionalIterator1 textEnd,
				      BidirectionalIterator2 pattern,
				      BidirectionalIterator2 patternEnd)
{
  return __search_L(text, textEnd, pattern, patternEnd, distance_type(text), distance_type(pattern));
}

// ------------------------------------------------------------------------

template <class ForwardIterator1, class ForwardIterator2, class Distance1,
	  class Distance2>
vector<ForwardIterator1>
__multi_QSsearch(ForwardIterator1 first1, ForwardIterator1 last1,
	       ForwardIterator2 first2, ForwardIterator2 last2,
	       Distance1*, Distance2*) {
    Distance1 d1 = 0;
    distance(first1, last1, d1);
    Distance2 d2 = 0;
    distance(first2, last2, d2);

    vector<ForwardIterator1> position;

    if (d1 < d2)
      return position;
 
    int
      disp,
      table[ASIZE];

    for (int i = 0; i < ASIZE; i++)
      table[i] = d2 + 1;

    for (ForwardIterator2 current = first2; current != last2; ++current)
      table[(int)(*current)] = d2 - (current - first2);

    ForwardIterator1 current1 = first1;
    ForwardIterator2 current2 = first2;
    
    while (1) { 
      while (current2 != last2)
	if (*current1++ != *current2++) {
	  if (d1 <= d2)
	    return position;
	  disp = table[(int)(*(first1 + d2))];
	  d1 -= disp;
	  first1  += disp;
	  current1 = first1;
	  current2 = first2;
	}
      position.push_back(first1);
      if (d1 <= d2)
	return position;
      disp = table[(int)(*(first1 + d2))];
      d1 -= disp;
      first1  += disp;
      current1 = first1;
      current2 = first2;
    }
}


template <class ForwardIterator1, class ForwardIterator2>
inline vector<ForwardIterator1>
multi_QSsearch(ForwardIterator1 first1, ForwardIterator1 last1,
	     ForwardIterator2 first2, ForwardIterator2 last2)
{
    return __multi_QSsearch(first1, last1, first2, last2,
			  distance_type(first1), distance_type(first2));
}

// ------------------------------------------------------------------------

template <class RandomAccessIterator1, class RandomAccessIterator2>
inline RandomAccessIterator1 Hsearch(RandomAccessIterator1 text,
				     RandomAccessIterator1 textEnd,
				     RandomAccessIterator2 pattern,
				     RandomAccessIterator2 patternEnd) {
  return __search_hashed(text, textEnd, pattern, patternEnd, distance_type(text), distance_type(pattern)); 
}


template <class RandomAccessIterator1, class RandomAccessIterator2, class Distance1, class Distance2>
RandomAccessIterator1 __search_hashed(RandomAccessIterator1 text,
                                    RandomAccessIterator1 textEnd,
                                    RandomAccessIterator2 pattern,
                                    RandomAccessIterator2 patternEnd,
				    Distance1*, Distance2*) {
  if (pattern == patternEnd)
    return text;
  Distance2 pattern_size, j, m;
  pattern_size = patternEnd - pattern; 
  if (suffix_size == 0 || pattern_size < suffix_size)
    return __search_L(text, textEnd, pattern, patternEnd, distance_type(text), distance_type(pattern));
  Distance1 i, k, large, adjustment, mismatch_shift, text_size;
  vector<Distance1> next, skip;
  k = 0; 
  text_size = textEnd - text;
  compute_next(pattern, patternEnd, next);
  if (next.size() == 1)
    return find(text, textEnd, *pattern);
  m = next.size();
  for (i = 0; i < hash_range_max; ++i)
    skip.push_back(m - suffix_size + 1);
  for (j = suffix_size - 1; j < m - 1; ++j)
    skip[hash(pattern + j)] = m - 1 - j;
  mismatch_shift = skip[hash(pattern + m - 1)];
  skip[hash(pattern + m - 1)] = 0;
  large = text_size + 1;
  adjustment = large + pattern_size - 1;
  skip[hash(pattern + pattern_size - 1)] = large;
  k -= text_size;
  for(;;) {
    k += pattern_size - 1;
    if (k >= 0) break;
    do {
      k += skip[hash(textEnd + k)]; 
    } while (k < 0);
    if (k < pattern_size)
      return textEnd;
    k -= adjustment;
    if (textEnd[k] != pattern[0])
      k += mismatch_shift;
    else {
      j = 1; 
      for (;;) {
        ++k;
        if (textEnd[k] != pattern[j])
          break;
        ++j;
        if (j == pattern_size)
          return textEnd + k - pattern_size + 1;
      }
      if (mismatch_shift > j)
        k += mismatch_shift - j;
      else
        for (;;) {
          j = next[j];
          if (j < 0) {
            ++k;
            break;
          }
          if (j == 0)
            break;
          while (textEnd[k] == pattern[j]) {
            ++k; ++j;
            if (j == pattern_size) {
              return textEnd + k - pattern_size;
            }
            if (k == 0)
              return textEnd;
          }
        }
    }
  }
  return textEnd;
}

#endif
